@@ -313,7 +313,8 @@ nsXBLPrototypeBinding::nsXBLPrototypeBinding()
 nsresult
 nsXBLPrototypeBinding::Init(const nsACString& aID,
                             nsIXBLDocumentInfo* aInfo,
-                            nsIContent* aElement)
+                            nsIContent* aElement,
+                            PRBool aFirstBinding)
 {
   if (!kAttrPool || !nsXBLInsertionPointEntry::PoolInited()) {
     return NS_ERROR_OUT_OF_MEMORY;
@@ -325,8 +326,13 @@ nsXBLPrototypeBinding::Init(const nsACString& aID,
   // The binding URI might not be a nsIURL (e.g. for data: URIs). In that case,
   // we always use the first binding, so we don't need to keep track of the ID.
   nsCOMPtr<nsIURL> bindingURL = do_QueryInterface(mBindingURI);
-  if (bindingURL)
+  if (bindingURL) {
+    if (aFirstBinding) {
+      rv = mBindingURI->Clone(getter_AddRefs(mAlternateBindingURI));
+      NS_ENSURE_SUCCESS(rv, rv);
+    }
     bindingURL->SetRef(aID);
+  }
 
   mXBLDocInfoWeak = aInfo;
 
@@ -334,6 +340,16 @@ nsXBLPrototypeBinding::Init(const nsACString& aID,
   return NS_OK;
 }
 
+PRBool nsXBLPrototypeBinding::CompareBindingURI(nsIURI* aURI) const
+{
+  PRBool equal = PR_FALSE;
+  mBindingURI->Equals(aURI, &equal);
+  if (!equal && mAlternateBindingURI) {
+    mAlternateBindingURI->Equals(aURI, &equal);
+  }
+  return equal;
+}
+
 static PRIntn
 TraverseInsertionPoint(nsHashKey* aKey, void* aData, void* aClosure)
 {
@@ -1057,23 +1073,6 @@ nsXBLPrototypeBinding::GetStyleSheets()
   return nsnull;
 }
 
-PRBool
-nsXBLPrototypeBinding::ShouldBuildChildFrames() const
-{
-  if (!mAttributeTable)
-    return PR_TRUE;
-  nsPRUint32Key nskey(kNameSpaceID_XBL);
-  nsObjectHashtable* xblAttributes =
-    static_cast<nsObjectHashtable*>(mAttributeTable->Get(&nskey));
-  if (xblAttributes) {
-    nsISupportsKey key(nsGkAtoms::text);
-    void* entry = xblAttributes->Get(&key);
-    return !entry;
-  }
-
-  return PR_TRUE;
-}
-
 static PRBool
 DeleteAttributeEntry(nsHashKey* aKey, void* aData, void* aClosure)
 {
