@@ -1178,16 +1178,16 @@ gfxWindowsFont::Measure(gfxTextRun *aTextRun,
     // we need to create a copy in order to avoid getting cached extents
     if (aBoundingBoxType == TIGHT_HINTED_OUTLINE_EXTENTS &&
         mAntialiasOption != CAIRO_ANTIALIAS_NONE) {
-        // Not nsRefPtr here as we know this is a transient font instance,
-        // and we won't be putting it in the font cache. So we want to
-        // delete it immediately it goes out of scope, not call
-        // gfxFont::Release which deals with shared, cached font instances.
-        nsAutoPtr<gfxWindowsFont> tempFont =
-            new gfxWindowsFont(GetFontEntry(), GetStyle(), CAIRO_ANTIALIAS_NONE);
-        if (tempFont) {
-            return tempFont->Measure(aTextRun, aStart, aEnd,
-                                     TIGHT_HINTED_OUTLINE_EXTENTS,
-                                     aRefContext, aSpacing);
+        // the non-antialiased font is held in an nsAutoPtr, so it will
+        // be destroyed when the "parent" font is flushed from the cache
+        if (!mNonAAFont) {
+            mNonAAFont = new gfxWindowsFont(GetFontEntry(), GetStyle(),
+                                            CAIRO_ANTIALIAS_NONE);
+        }
+        if (mNonAAFont) {
+            return mNonAAFont->Measure(aTextRun, aStart, aEnd,
+                                       TIGHT_HINTED_OUTLINE_EXTENTS,
+                                       aRefContext, aSpacing);
         }
     }
 
@@ -1536,9 +1536,9 @@ gfxWindowsFontGroup::MakeTextRun(const PRUint8 *aString, PRUint32 aLength,
         nsAutoString utf16;
         AppendASCIItoUTF16(cString, utf16);
         if (isComplex) {
-            InitTextRunUniscribe(aParams->mContext, textRun, utf16.get(), aLength);
+            InitTextRunUniscribe(aParams->mContext, textRun, utf16.get(), utf16.Length());
         } else {
-            InitTextRunGDI(aParams->mContext, textRun, utf16.get(), aLength);
+            InitTextRunGDI(aParams->mContext, textRun, utf16.get(), utf16.Length());
         }
     }
 
@@ -1825,15 +1825,18 @@ public:
         mFontSelected(PR_FALSE), mForceGDIPlace(PR_FALSE)
     {
         NS_ASSERTION(mMaxGlyphs < 65535, "UniscribeItem is too big, ScriptShape() will fail!");
-        mGlyphs.SetLength(mMaxGlyphs);
-        mClusters.SetLength(mItemLength + 1);
-        mAttr.SetLength(mMaxGlyphs);
     }
 
     ~UniscribeItem() {
         free(mAlternativeString);
     }
 
+    PRBool AllocateBuffers() {
+        return (mGlyphs.SetLength(mMaxGlyphs) &&
+                mClusters.SetLength(mItemLength + 1) &&
+                mAttr.SetLength(mMaxGlyphs));
+    }
+
     /* possible return values:
      * S_OK - things succeeded
      * GDI_ERROR - things failed to shape.  Might want to try again after calling DisableShaping()
@@ -1870,8 +1873,10 @@ public:
 
             if (rv == E_OUTOFMEMORY) {
                 mMaxGlyphs *= 2;
-                mGlyphs.SetLength(mMaxGlyphs);
-                mAttr.SetLength(mMaxGlyphs);
+                if (!mGlyphs.SetLength(mMaxGlyphs) ||
+                    !mAttr.SetLength(mMaxGlyphs)) {
+                    return E_OUTOFMEMORY;
+                }
                 continue;
             }
 
@@ -2009,8 +2014,10 @@ public:
     }
 
     HRESULT Place() {
-        mOffsets.SetLength(mNumGlyphs);
-        mAdvances.SetLength(mNumGlyphs);
+        if (!mOffsets.SetLength(mNumGlyphs) ||
+            !mAdvances.SetLength(mNumGlyphs)) {
+            return E_OUTOFMEMORY;
+        }
 
         if (mForceGDIPlace)
             return PlaceGDI();
@@ -2267,9 +2274,13 @@ private:
 };
 
 
-#define MAX_ITEM_LENGTH 32768
-
+// This is _slightly less_ than half of the maximum analysis window
+// we use with ScriptBreak, so that in typical cases we end up re-processing
+// only a small amount of overlap when we move the analysis window forward.
+#define MAX_ITEM_LENGTH      32499
 
+// Limit length of text passed to Uniscribe APIs, to avoid failure there.
+#define MAX_UNISCRIBE_LENGTH 65000
 
 static PRUint32 FindNextItemStart(int aOffset, int aLimit,
                                   nsTArray<SCRIPT_LOGATTR> &aLogAttr,
@@ -2332,28 +2343,53 @@ private:
     nsresult CopyItemSplitOversize(int aIndex, nsTArray<SCRIPT_ITEM> &aDest) {
         aDest.AppendElement(mItems[aIndex]);
         const int itemLength = mItems[aIndex+1].iCharPos - mItems[aIndex].iCharPos;
-        if (ESTIMATE_MAX_GLYPHS(itemLength) > 65535) {
+        if (ESTIMATE_MAX_GLYPHS(itemLength) > MAX_UNISCRIBE_LENGTH) {
             // This items length would cause ScriptShape() to fail. We need to
             // add extra items here so that no item's length could cause the fail.
 
             // Get cluster boundaries, so we can break cleanly if possible.
             nsTArray<SCRIPT_LOGATTR> logAttr;
-            if (!logAttr.SetLength(itemLength))
-                return NS_ERROR_FAILURE;
-            HRESULT rv= ScriptBreak(mString+mItems[aIndex].iCharPos, itemLength,
-                                    &mItems[aIndex].a, logAttr.Elements());
-            if (FAILED(rv))
-                return NS_ERROR_FAILURE;
 
             const int nextItemStart = mItems[aIndex+1].iCharPos;
-            int start = FindNextItemStart(mItems[aIndex].iCharPos,
-                                          nextItemStart, logAttr, mString);
+            int start = mItems[aIndex].iCharPos;
 
             while (start < nextItemStart) {
-                SCRIPT_ITEM item = mItems[aIndex];
-                item.iCharPos = start;
-                aDest.AppendElement(item);
-                start = FindNextItemStart(start, nextItemStart, logAttr, mString);
+                // ScriptBreak will fail for strings longer than 64K,
+                // so we do the analysis using a "sliding window" over the
+                // huge item.
+                int analysisLen = PR_MIN(nextItemStart - start,
+                                         MAX_UNISCRIBE_LENGTH);
+                if (!logAttr.SetLength(analysisLen)) {
+                    return NS_ERROR_FAILURE;
+                }
+                HRESULT rv = ScriptBreak(mString + start,
+                                         analysisLen,
+                                         &mItems[aIndex].a,
+                                         logAttr.Elements());
+                if (FAILED(rv)) {
+                    return NS_ERROR_FAILURE;
+                }
+
+                int analysisLimit = start + analysisLen;
+                start = FindNextItemStart(start, analysisLimit,
+                                          logAttr, mString);
+                int prevStart = start;
+                while (start < analysisLimit) {
+                    SCRIPT_ITEM item = mItems[aIndex];
+                    item.iCharPos = start;
+                    aDest.AppendElement(item);
+                    prevStart = start;
+                    start = FindNextItemStart(start, analysisLimit,
+                                              logAttr, mString);
+                }
+
+                // If the analysis window didn't reach the end of the entire
+                // original item, reset start so that the final (perhaps
+                // badly-terminated) item we just created will be merged with
+                // the following section of the text.
+                if (start < nextItemStart) {
+                    start = prevStart;
+                }
             }
         } 
         return NS_OK;
@@ -2382,7 +2418,7 @@ public:
             Init();
         }
 
-        if (ESTIMATE_MAX_GLYPHS(mLength) > 65535) {
+        if (ESTIMATE_MAX_GLYPHS(mLength) > MAX_UNISCRIBE_LENGTH) {
             // Any item of length > 43680 will cause ScriptShape() to fail, as its
             // mMaxGlyphs value will be greater than 65535 (43680*1.5+16>65535). So we
             // need to break up items which are longer than that upon cluster boundaries.
@@ -2414,7 +2450,10 @@ public:
                                                 mItems[i+1].iCharPos - mItems[i].iCharPos,
                                                 &mItems[i],
                                                 aGroup);
-
+        if (!item->AllocateBuffers()) {
+            delete item;
+            return nsnull;
+        }
         return item;
     }
 
@@ -2646,9 +2685,13 @@ gfxWindowsFontGroup::InitTextRunUniscribe(gfxContext *aContext, gfxTextRun *aRun
     int numItems = us.Itemize();
 
     for (int i = 0; i < numItems; ++i) {
-        SaveDC(aDC);
-
         nsAutoPtr<UniscribeItem> item(us.GetItem(i, this));
+        if (!item) {
+            // failed to initialize the item; give up (out of memory)
+            break;
+        }
+
+        SaveDC(aDC);
 
         // jtdfix - push this into the pref handling code??
         mItemLangGroup = nsnull;
