@@ -35,10 +35,10 @@
  *
  * ***** END LICENSE BLOCK ***** */
 
-var gTestfile = 'regress-360969-05.js';
+var gTestfile = 'regress-360969-03.js';
 //-----------------------------------------------------------------------------
 var BUGNUMBER = 360969;
-var summary = '2^17: local function';
+var summary = '2^17: local const';
 var actual = 'No Crash';
 var expect = 'No Crash';
 
@@ -46,7 +46,6 @@ var global = this;
 
 //-----------------------------------------------------------------------------
 test();
-
 //-----------------------------------------------------------------------------
 
 function test()
@@ -62,7 +61,7 @@ function test()
 
   for (var i = 0; i < limit; i++)
   {
-    eval('function pf' + i + '() {}');
+    eval('const pv' + i + ';');
   }
 
   reportCompare(expect, actual, summary);
@@ -73,4 +72,3 @@ function test()
 
   exitFunc ('test');
 }
-
