@@ -16,10 +16,10 @@
  *
  * The Initial Developer of the Original Code is
  * Mozilla Foundation.
- * Portions created by the Initial Developer are Copyright (C) 2007
+ * Portions created by the Initial Developer are Copyright (C) 2008
  * the Initial Developer. All Rights Reserved.
  *
- * Contributor(s): Jesse Ruderman
+ * Contributor(s): Igor Bukanov
  *
  * Alternatively, the contents of this file may be used under the terms of
  * either the GNU General Public License Version 2 or later (the "GPL"), or
@@ -35,13 +35,13 @@
  *
  * ***** END LICENSE BLOCK ***** */
 
-var gTestfile = 'regress-385393-01.js';
-
+var gTestfile = 'regress-369696-03.js';
 //-----------------------------------------------------------------------------
-var BUGNUMBER = 385393;
-var summary = 'Regression test for bug 385393';
-var actual = 'No Crash';
-var expect = 'No Crash';
+var BUGNUMBER = 396696;
+var summary = 'Do not assert: map->depth > 0" in js_LeaveSharpObject';
+var actual = '';
+var expect = '';
+
 
 //-----------------------------------------------------------------------------
 test();
@@ -53,13 +53,26 @@ function test()
   printBugNumber(BUGNUMBER);
   printStatus (summary);
 
-  try
-  {
-    [].map(1 for (x in []));
-  }
-  catch(ex)
-  {
-  }
+  var x = [[[ { toSource: function() { gc();  }}]]];
+
+  var a = [];
+  a[0] = a;
+  a.toSource = a.toString;
+  Array.prototype.toSource.call(a);
+
+//cx->sharpObjectMap.depth == -2
+
+  (function() {
+    var tmp = [];
+    for (var i = 0; i != 30*1000; ++i) {
+      var tmp2 = [];
+      tmp.push(tmp2);
+      tmp2.toSource();
+    }
+  })();
+
+  gc();
+  x.toSource();
 
   reportCompare(expect, actual, summary);
 
