@@ -16,10 +16,10 @@
  *
  * The Initial Developer of the Original Code is
  * Mozilla Foundation.
- * Portions created by the Initial Developer are Copyright (C) 2008
+ * Portions created by the Initial Developer are Copyright (C) 2007
  * the Initial Developer. All Rights Reserved.
  *
- * Contributor(s):
+ * Contributor(s): Jesse Ruderman
  *
  * Alternatively, the contents of this file may be used under the terms of
  * either the GNU General Public License Version 2 or later (the "GPL"), or
@@ -35,10 +35,10 @@
  *
  * ***** END LICENSE BLOCK ***** */
 
-var gTestfile = 'regress-452498-110.js';
+var gTestfile = 'regress-373827-02.js';
 //-----------------------------------------------------------------------------
-var BUGNUMBER = 452498;
-var summary = 'TM: upvar2 regression tests';
+var BUGNUMBER = 373827;
+var summary = 'Do not assert: OBJ_GET_CLASS(cx, obj)->flags & JSCLASS_HAS_PRIVATE';
 var actual = '';
 var expect = '';
 
@@ -52,31 +52,16 @@ function test()
   enterFunc ('test');
   printBugNumber(BUGNUMBER);
   printStatus (summary);
+ 
+  try
+  {
+    let ([] = [{x: function(){}}]) { }; foo;
+  }
+  catch(ex)
+  {
+  }
 
-// ------- Comment #110 From Brendan Eich
-
-// (In reply to comment #107)
-
-  function f(a) { const b = a; print(++b); return b; }
-
-  expect = 'function f(a) { const b = a; print(+ b + 1); return b; }';
-  actual = f + '';
-  compareSource(expect, actual, 'function f(a) { const b = a; print(++b); return b; }');
-
-  expect = '01';
-  actual = 0;
-
-  function g(a) { const b = a; print(actual = ++b); return b; }
-  actual = String(actual) + g(1);
-  reportCompare(expect, actual, 'function g(a) { const b = a; print(actual = ++b); return b; }');
-
-  expect = '21';
-  actual = 0;
-
-  const x = 1; print(actual = ++x);
-  actual = String(actual) + x;
-
-  reportCompare(expect, actual, 'const x = 1; print(actual = ++x); ');
+  reportCompare(expect, actual, summary);
 
   exitFunc ('test');
 }
