@@ -36,6 +36,10 @@
  *
  * ***** END LICENSE BLOCK ***** */
 
+#ifdef MOZ_IPC
+#include "base/basictypes.h"
+#endif
+
 #include "nsJSNPRuntime.h"
 #include "nsNPAPIPlugin.h"
 #include "nsNPAPIPluginInstance.h"
@@ -53,6 +57,14 @@
 // FIXME(bug 332648): Give me a real API please!
 #include "jscntxt.h"
 
+using namespace mozilla::plugins::parent;
+
+#ifdef MOZ_IPC
+#include "mozilla/plugins/PluginScriptableObjectParent.h"
+using mozilla::plugins::PluginScriptableObjectParent;
+using mozilla::plugins::ParentNPObject;
+#endif
+
 // Hash of JSObject wrappers that wraps JSObjects as NPObjects. There
 // will be one wrapper per JSObject per plugin instance, i.e. if two
 // plugins access the JSObject x, two wrappers for x will be
@@ -78,6 +90,21 @@ static JSRuntime *sJSRuntime;
 // while executing JS on the context.
 static nsIJSContextStack *sContextStack;
 
+static nsTArray<NPObject*>* sDelayedReleases;
+
+namespace {
+
+inline bool
+NPObjectIsOutOfProcessProxy(NPObject *obj)
+{
+#ifdef MOZ_IPC
+  return obj->_class == PluginScriptableObjectParent::GetClass();
+#else
+  return false;
+#endif
+}
+
+} // anonymous namespace
 
 // Helper class that reports any JS exceptions that were thrown while
 // the plugin executed JS.
@@ -152,8 +179,8 @@ NPObjWrapper_Construct(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
                        jsval *rval);
 
 static JSBool
-CreateNPObjectMember(NPP npp, JSContext *cx, JSObject *obj,
-                     NPObject *npobj, jsval id, jsval *vp);
+CreateNPObjectMember(NPP npp, JSContext *cx, JSObject *obj, NPObject *npobj,
+                     jsval id, NPVariant* getPropertyResult, jsval *vp);
 
 static JSClass sNPObjectJSWrapperClass =
   {
@@ -197,18 +224,46 @@ static JSClass sNPObjectMemberClass =
     nsnull, nsnull, nsnull, NPObjectMember_Mark, nsnull
   };
 
+static void
+OnWrapperDestroyed();
+
+static JSBool
+DelayedReleaseGCCallback(JSContext* cx, JSGCStatus status)
+{
+  if (JSGC_END == status) {
+    // Take ownership of sDelayedReleases and null it out now. The
+    // _releaseobject call below can reenter GC and double-free these objects.
+    nsAutoPtr<nsTArray<NPObject*> > delayedReleases(sDelayedReleases);
+    sDelayedReleases = nsnull;
+
+    if (delayedReleases) {
+      for (PRUint32 i = 0; i < delayedReleases->Length(); ++i) {
+        NPObject* obj = (*delayedReleases)[i];
+        if (obj)
+          _releaseobject(obj);
+        OnWrapperDestroyed();
+      }
+    }
+  }
+  return JS_TRUE;
+}
+
 static void
 OnWrapperCreated()
 {
   if (sWrapperCount++ == 0) {
     static const char rtsvc_id[] = "@mozilla.org/js/xpc/RuntimeService;1";
-    nsCOMPtr<nsIJSRuntimeService> rtsvc(do_GetService(rtsvc_id));
+    nsCOMPtr<nsIJSRuntimeService_MOZILLA_1_9_2> rtsvc(do_GetService(rtsvc_id));
     if (!rtsvc)
       return;
 
     rtsvc->GetRuntime(&sJSRuntime);
     NS_ASSERTION(sJSRuntime != nsnull, "no JSRuntime?!");
 
+    // Register our GC callback to perform delayed destruction of finalized
+    // NPObjects. Leave this callback around and don't ever unregister it.
+    rtsvc->RegisterGCCallback(DelayedReleaseGCCallback);
+
     CallGetService("@mozilla.org/js/xpc/ContextStack;1", &sContextStack);
   }
 }
@@ -488,11 +543,14 @@ ReportExceptionIfPending(JSContext *cx)
 nsJSObjWrapper::nsJSObjWrapper(NPP npp)
   : nsJSObjWrapperKey(nsnull, npp)
 {
+  MOZ_COUNT_CTOR(nsJSObjWrapper);
   OnWrapperCreated();
 }
 
 nsJSObjWrapper::~nsJSObjWrapper()
 {
+  MOZ_COUNT_DTOR(nsJSObjWrapper);
+
   // Invalidate first, since it relies on sJSRuntime and sJSObjWrappers.
   NP_Invalidate(this);
 
@@ -1041,7 +1099,8 @@ nsJSObjWrapper::GetNewOrUsed(NPP npp, JSContext *cx, JSObject *obj)
 
     NPObject *npobj = (NPObject *)::JS_GetPrivate(cx, obj);
 
-    return _retainobject(npobj);
+    if (LookupNPP(npobj) == npp)
+      return _retainobject(npobj);
   }
 
   if (!sJSObjWrappers.ops) {
@@ -1147,6 +1206,10 @@ NPObjWrapper_AddProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
     return JS_FALSE;
   }
 
+  if (NPObjectIsOutOfProcessProxy(npobj)) {
+    return JS_TRUE;
+  }
+
   PluginDestructionGuard pdg(LookupNPP(npobj));
 
   JSBool hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
@@ -1185,12 +1248,14 @@ NPObjWrapper_DelProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
 
   PluginDestructionGuard pdg(LookupNPP(npobj));
 
-  JSBool hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
-  if (!ReportExceptionIfPending(cx))
-    return JS_FALSE;
+  if (!NPObjectIsOutOfProcessProxy(npobj)) {
+    JSBool hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
+    if (!ReportExceptionIfPending(cx))
+      return JS_FALSE;
 
-  if (!hasProperty)
-    return JS_TRUE;
+    if (!hasProperty)
+      return JS_TRUE;
+  }
 
   if (!npobj->_class->removeProperty(npobj, (NPIdentifier)id))
     *vp = JSVAL_FALSE;
@@ -1222,14 +1287,16 @@ NPObjWrapper_SetProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
 
   PluginDestructionGuard pdg(npp);
 
-  JSBool hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
-  if (!ReportExceptionIfPending(cx))
-    return JS_FALSE;
+  if (!NPObjectIsOutOfProcessProxy(npobj)) {
+    JSBool hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
+    if (!ReportExceptionIfPending(cx))
+      return JS_FALSE;
 
-  if (!hasProperty) {
-    ThrowJSException(cx, "Trying to set unsupported property on NPObject!");
+    if (!hasProperty) {
+      ThrowJSException(cx, "Trying to set unsupported property on NPObject!");
 
-    return JS_FALSE;
+      return JS_FALSE;
+    }
   }
 
   NPVariant npv;
@@ -1276,22 +1343,58 @@ NPObjWrapper_GetProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
 
   PluginDestructionGuard pdg(npp);
 
-  PRBool hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
+  PRBool hasProperty, hasMethod;
+
+  NPVariant npv;
+  VOID_TO_NPVARIANT(npv);
+
+#ifdef MOZ_IPC
+  if (NPObjectIsOutOfProcessProxy(npobj)) {
+    PluginScriptableObjectParent* actor =
+      static_cast<ParentNPObject*>(npobj)->parent;
+
+    // actor may be null if the plugin crashed.
+    if (!actor)
+      return JS_FALSE;
+
+    JSBool success = actor->GetPropertyHelper((NPIdentifier)id, &hasProperty,
+                                              &hasMethod, &npv);
+    if (!ReportExceptionIfPending(cx)) {
+      if (success)
+        _releasevariantvalue(&npv);
+      return JS_FALSE;
+    }
+
+    if (success) {
+      // We return NPObject Member class here to support ambiguous members.
+      if (hasProperty && hasMethod)
+        return CreateNPObjectMember(npp, cx, obj, npobj, id, &npv, vp);
+
+      if (hasProperty) {
+        *vp = NPVariantToJSVal(npp, cx, &npv);
+        _releasevariantvalue(&npv);
+
+        if (!ReportExceptionIfPending(cx))
+          return JS_FALSE;
+      }
+    }
+    return JS_TRUE;
+  }
+#endif
+
+  hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
   if (!ReportExceptionIfPending(cx))
     return JS_FALSE;
 
-  PRBool hasMethod = npobj->_class->hasMethod(npobj, (NPIdentifier)id);
+  hasMethod = npobj->_class->hasMethod(npobj, (NPIdentifier)id);
   if (!ReportExceptionIfPending(cx))
     return JS_FALSE;
 
   // We return NPObject Member class here to support ambiguous members.
   if (hasProperty && hasMethod)
-    return CreateNPObjectMember(npp, cx, obj, npobj, id, vp);
+    return CreateNPObjectMember(npp, cx, obj, npobj, id, nsnull, vp);
 
   if (hasProperty) {
-    NPVariant npv;
-    VOID_TO_NPVARIANT(npv);
-
     if (npobj->_class->getProperty(npobj, (NPIdentifier)id, &npv))
       *vp = NPVariantToJSVal(npp, cx, &npv);
 
@@ -1625,17 +1728,15 @@ static void
 NPObjWrapper_Finalize(JSContext *cx, JSObject *obj)
 {
   NPObject *npobj = (NPObject *)::JS_GetPrivate(cx, obj);
-
   if (npobj) {
     if (sNPObjWrappers.ops) {
       PL_DHashTableOperate(&sNPObjWrappers, npobj, PL_DHASH_REMOVE);
     }
-
-    // Let go of our NPObject
-    _releaseobject(npobj);
   }
 
-  OnWrapperDestroyed();
+  if (!sDelayedReleases)
+    sDelayedReleases = new nsTArray<NPObject*>;
+  sDelayedReleases->AppendElement(npobj);
 }
 
 static JSBool
@@ -1856,6 +1957,16 @@ NPObjWrapperPluginDestroyedCallback(PLDHashTable *table, PLDHashEntryHdr *hdr,
       npobj->_class->invalidate(npobj);
     }
 
+#ifdef NS_BUILD_REFCNT_LOGGING
+    {
+      int32_t refCnt = npobj->referenceCount;
+      while (refCnt) {
+        --refCnt;
+        NS_LOG_RELEASE(npobj, refCnt, "BrowserNPObject");
+      }
+    }
+#endif
+
     // Force deallocation of plugin objects since the plugin they came
     // from is being torn down.
     if (npobj->_class && npobj->_class->deallocate) {
@@ -1988,10 +2099,8 @@ static NPP
 LookupNPP(NPObject *npobj)
 {
   if (npobj->_class == &nsJSObjWrapper::sJSObjWrapperNPClass) {
-    NS_ERROR("NPP requested for NPObject of class "
-             "nsJSObjWrapper::sJSObjWrapperNPClass!\n");
-
-    return nsnull;
+    nsJSObjWrapper* o = static_cast<nsJSObjWrapper*>(npobj);
+    return o->mNpp;
   }
 
   NPObjWrapperHashEntry *entry = static_cast<NPObjWrapperHashEntry *>
@@ -2007,8 +2116,8 @@ LookupNPP(NPObject *npobj)
 }
 
 JSBool
-CreateNPObjectMember(NPP npp, JSContext *cx, JSObject *obj,
-                     NPObject* npobj, jsval id, jsval *vp)
+CreateNPObjectMember(NPP npp, JSContext *cx, JSObject *obj, NPObject* npobj,
+                     jsval id,  NPVariant* getPropertyResult, jsval *vp)
 {
   NS_ENSURE_TRUE(vp, JS_FALSE);
 
@@ -2041,18 +2150,27 @@ CreateNPObjectMember(NPP npp, JSContext *cx, JSObject *obj,
 
   jsval fieldValue;
   NPVariant npv;
-  VOID_TO_NPVARIANT(npv);
+  NPBool hasProperty;
 
-  NPBool hasProperty = npobj->_class->getProperty(npobj, (NPIdentifier)id,
-                                                  &npv);
-  if (ReportExceptionIfPending(cx)) {
-    ::JS_RemoveRoot(cx, vp);
-    return JS_FALSE;
+  if (getPropertyResult) {
+    // Plugin has already handed us the value we want here.
+    npv = *getPropertyResult;
+    hasProperty = true;
   }
+  else {
+    VOID_TO_NPVARIANT(npv);
 
-  if (!hasProperty) {
-    ::JS_RemoveRoot(cx, vp);
-    return JS_FALSE;
+    NPBool hasProperty = npobj->_class->getProperty(npobj, (NPIdentifier)id,
+                                                    &npv);
+    if (!ReportExceptionIfPending(cx)) {
+      ::JS_RemoveRoot(cx, vp);
+      return JS_FALSE;
+    }
+
+    if (!hasProperty) {
+      ::JS_RemoveRoot(cx, vp);
+      return JS_FALSE;
+    }
   }
 
   fieldValue = NPVariantToJSVal(npp, cx, &npv);
